<?php
/**
 * UpdateTableNS.php
 * Rebuilds the wikidb_tabledata table, which holds the parsed field definitions for
 * all pages located in a table namespace (as defined by $wgWikiDBNamespaces).
 *
 * You should run this script whenever you add or remove a table namespace (i.e.
 * when you modify $wgWikiDBNamespaces).  Call it with the IDs of any namespaces
 * that have been added or removed from the array, separated by spaces.
 * You can specify 'all' to rebuild all table namespaces, but this is slow and
 * should only be necessary if an upgrade requires it.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2017, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 1341 $
 */

$Usage = <<<EOF
Specify the namespaces you want to update using their namespace IDs.
Separate namespaces with a space.
Alternatively, use 'all' (without quotes) to update all namespaces.
The --skip-refresh argument can be used to bypass the code which updates
any stale records created due to data definitions changing.

EOF;
//////////////////////////////////////

// Include header file required for all WikiDB maintenance scripts.
	require_once(dirname( __FILE__ ) . "/header.inc");

// Check the command-line arguments to get the list of namespaces to check.
	$arrNameSpaces = array();
	$TruncateTableFirst = false;

	foreach ($args as $NameSpace) {
	// If 'all' specified set a flag to indicate that the whole table should be
	// wiped, and then add the table namespaces (which need rebuilding) to the array.
		if (strtolower($NameSpace) == "all") {
			$TruncateTableFirst = true;
			$arrNameSpaces = array_keys($wgWikiDBNamespaces);
		// Break out of loop - no point looking at other arguments, as all namespaces
		// are already being updated.
			break;
		}

	// If an argument is not a valid positive integer then die with an error.
		if (!preg_match('/^\d+$/', $NameSpace))
			die("Invalid namespace ID: " . $NameSpace . "\n");

	// Otherwise convert it to an integer and add it to our array of namespaces to
	// rebuild.
		$arrNameSpaces[] = intval($NameSpace);
	}

// If no valid arguments specified, display usage instructions and die.
	if (count($arrNameSpaces) == 0)
		die($Usage);

// Tidy up the input: remove any duplicates, and sort the namespaces in
// numerical order.
	$arrNameSpaces = array_unique($arrNameSpaces);
	sort($arrNameSpaces);

// Give a warning, as this can be a very slow script on large wikis.
	print("About to update cached table definitions for ");
	if ($TruncateTableFirst)
		print("ALL namespaces");
	else {
		print(count($arrNameSpaces) . " namespace");
		if (count($arrNameSpaces) > 1)
			print("s");
	}
	print(".\nWarning: This can be a very slow process for large databases!\n");
	AllowUserToCancel();

// If a full rebuild is requested, wipe the table first.
	if ($TruncateTableFirst) {
		print("'all' specified - clearing cache completely...");
		$AffectedRows = ClearTableDefCache();
		ReportResult("deleted", $AffectedRows);
	}

	foreach ($arrNameSpaces as $NS) {
		if (!isset($wgWikiDBNamespaces[$NS])) {
		// Don't need to check $TruncateTableFirst, as you can't get here if
		// that option has been set.
			ReportNSUpdate("Clearing", $NS);
			$AffectedRows = ClearTableDefCache($NS);
			ReportResult("deleted", $AffectedRows);
		}
		else {
			if (!$TruncateTableFirst) {
				ReportNSUpdate("Clearing", $NS);
				$AffectedRows = ClearTableDefCache($NS);
				ReportResult("deleted", $AffectedRows);
			}

			ReportNSUpdate("Rebuilding", $NS);
			$AffectedRows = RebuildTableDefCache($NS);
			ReportResult("cached", $AffectedRows);
		}
	}

// Update stale data rows.  If '--skip-refresh' option specified then we only
// report how many are stale.
	if (isset($options['skip-refresh'])) {
		print("Fixing of stale records skipped: ");
		$options['report'] = true;
	}
	include(dirname(__FILE__) . "/RefreshStaleData.php");

// End of code - remainder of file is just helper functions.
	return;

///////////////// HELPER FUNCTIONS

// Based on refreshLinks()
// function refreshLinks($newOnly = false, $maxLag = false, $end = 0 )
function RebuildTableDefCache($NS) {
	global $wgUser, $wgParser, $wgUseImageResize, $wgUseTidy;

	$fname = __FUNCTION__;

	wfWaitForSlaves(false);

	$dbr = wfGetDB(DB_SLAVE);

// Don't generate TeX PNGs (lack of a sensible current directory causes errors
// anyway).  Built-in Maths support was removed in MW 1.18, so check for existence
// of the constant before attempting to use it.
	if (defined("MW_MATH_SOURCE")) {
		$wgUser->setOption('math', MW_MATH_SOURCE);
	}

// Don't generate thumbnail images
	$wgUseImageResize = false;
	$wgUseTidy = false;

	$res = $dbr->select('page', array('page_id'),
						array('page_namespace' => $NS), $fname);

	$i = 0;
	$LastReport = "";
	while ($row = $dbr->fetchObject($res)) {
		if ((++$i % pWIKIDB_ReportingInterval) == 0) {
			print_c_fixed($LastReport, $i);
			$LastReport = $i;
		}

		ReparseArticle($row->page_id);
	}

	print_c_fixed($LastReport, "");
	return $i;
}

///// REPORTING

function ReportNSUpdate($Action, $NS) {
	$NSName = WikiDB_GetCanonicalNamespaceName($NS);

// Remove brackets, if present (as may be the case for 'main') because we don't want
// a double set of brackets.
	if (substr($NSName, 0, 1) == "(" && substr($NSName, -1) == ")") {
		$NSName = substr($NSName, 1, -1);
	}

	print($Action . " NS-" . $NS . " (" . $NSName . ")...");
}

function ReportResult($Action, $AffectedRows) {
	print($AffectedRows . " table def");
	if ($AffectedRows != 1)
		print("s");
	print(" " . $Action . "\n");
}
